from read_write import read_pdb, write_pdb
from graph import SimpleGraph, GraphComporator
import numpy as np
import argparse


def get_unique_graphs(graphs):

    unique_graphs = [graphs[0]]
    for g_1 in graphs[1:]:
        new = True
        for g_2 in unique_graphs:
            if len(GraphComporator(g_1, g_2).find_isomorphic_bijections(all=False)) != 0:
                new = False
                break
        if new:
            unique_graphs.append(g_1)
    return unique_graphs


def find_inter_edges(bijection, graph):

    inter_edges = []
    for v_1, v_2 in bijection:
        if v_1 != bijection[v_2][1] or v_1 == v_2:
            return inter_edges
    traversed = np.zeros(graph.vertices_count, dtype=int)
    for v_1, v_2 in bijection:
        if traversed[v_1] == 0 and traversed[v_2] == 0:
            traversed[[v_1, v_2]] = 1
            if v_2 in graph.adjacency_list[v_1]:
                if v_1 > v_2:
                    v_1, v_2 = v_2, v_1
                inter_edges.append((v_1, v_2))
    return inter_edges


def split_graph(graph):

    graph_halves = []
    for bijection in np.array(GraphComporator(graph, graph).find_isomorphic_bijections()):
        inter_edges = find_inter_edges(bijection, graph)
        if len(inter_edges) != 0:
            new_graph = graph.copy()
            new_graph.remove_edges(inter_edges)
            connected_parts = new_graph.find_connected_parts()
            if len(connected_parts) == 2:
                new_vertices = connected_parts[0]
                new_indexes = {v: i for i, v in enumerate(new_vertices)}
                new_edges = [(new_indexes[v_1], new_indexes[v_2]) for v_1, v_2 in new_graph.edges
                             if (new_indexes.get(v_1) is not None and new_indexes.get(v_2) is not None)]
                half = SimpleGraph(len(new_vertices), new_edges)
                if len(new_graph.labels) != 0:
                    half.labels = new_graph.labels[new_vertices]
                if len(new_graph.coordinates) != 0:
                    half.coordinates = new_graph.coordinates[new_vertices]
                graph_halves.append(half)
    return graph_halves


def main():

    parser = argparse.ArgumentParser(description="Tile spliter.")
    parser.add_argument("-i", "--input_file", type=str, default=None, help="path to input pdb file")
    args = parser.parse_args()
    file_name = args.input_file[:-4]
    try:
        symbols, coordinates, edges = read_pdb(file_name)
    except:
        print("Can not read " + file_name + '!')
        return None
    if len(edges) == 0:
        print("The atomic connectivity is not found!")
        return None
    if len(symbols) % 2 == 0:
        graph = SimpleGraph(len(symbols), edges, labels=symbols, coordinates=coordinates)
        graph.calc_adjacency_list()
        parts = len(graph.find_connected_parts())
        if len(graph.find_connected_parts()) != 1:
            print("There are " + str(parts) + " isolated parts. Please check the atomic connectivity!")
            return None
        tile_halves = split_graph(graph)
        if len(tile_halves) != 0:
            for i, h in enumerate(get_unique_graphs(tile_halves)):
                try:
                    write_pdb(file_name + '_h' + str(i + 1), h.labels, h.coordinates, h.edges)
                except:
                    print('Can not write the output file!')
        else:
            print("The tile is impossible to split since it has no required automorphism!")
    else:
        print("The tile is impossible to split since it has an odd number of atoms!")
    return None


if __name__ == "__main__":
    main()
